#!/usr/local/bin/perl -w	

# Author: Mohamed Uduman
# Genomics & Bioinformatics
# Fall 2006
# Final Project 
# USAGE: CGI script to be viewed from browser
# Requirements: 
# ./prosite.dat <---PROSITE file (FULL)
# ../image/ <---directory to which this script outputs the myimage.png
# ./styless.css <--- CSS
	
use strict;
use CGI qw(:standard :html3);
use IO::String;
use Bio::Seq;
use Bio::SeqIO;
use CGI::Carp 'fatalsToBrowser';
use Storable qw(retrieve);
use Bio::Graphics;
use Bio::SeqFeature::Generic;

#$CGI::POST_MAX = 1024;
$CGI::DISABLE_UPLOADS = 1;
my @parameters = ();


# Header of HTML
print header(),
	 start_html(-title=>'Mohamed Uduman. Genomics & Bioinformatics. Fall 2006.',
					  -author=>'mohamed.uduman@yale.edu',
					  -base=>'true',
					  -target=>'_self',
					  -meta=>{'keywords'=>'PROSITE motif',
								 'copyright'=>'Copyright 2006  Mohamed Uduman'},
					  -style=>{'src'=>'styles.css'},
					  -BGCOLOR=>'#E4EBCF');



# TOP PANE (static)
print (&get_html_top);

# MAIN PANE (dynamic)
if(param("page") eq "about"){
	 print (&get_html_about);
}
elsif(param("submit")){
	 print (&get_html_results);
}
else{
	 #To reduce processing time I have stored the hash of motifs as presistent storage.
	 #This is used to populated the drop-down list in the main page
	 my $motifs_list_ref=retrieve ("motifs_list.hash"); # retreving stored hash
	 print (&get_html_main($motifs_list_ref));
}

# BOTTOM PANE (static)
print (&get_html_bottom);
print end_html();











########################################################
################## SUBROUTINES #########################
########################################################

# The static top pane, which includes the menu on the left
sub get_html_top{
	 my $html_top = qq(
  <TABLE WIDTH=100% ALIGN=CENTER>
  <TR>

    <TD>
      &nbsp;
    </TD>
  </TR>

  <TR>
    <TD>
      <TABLE WIDTH=95% ALIGN=CENTER CLASS=mainTable>
      <TR>

        <TD>
          <TABLE WIDTH=100% ALIGN=CENTER CELLPADDING=10>
          <TR>
            <TD COLSPAN=3>
              <H1>Genomics & Bioinformatics: Programming Project</H1>
            </TD>
          </TR>
          <TR ALIGN=LEFT VALIGN=TOP>

            <TD WIDTH=20% VALIGN=TOP>
              <TABLE WIDTH=100% ALIGN=LEFT VALIGN=TOP CELLPADDING=5 class=menu>
              <TR BGCOLOR=#B8CA80 VALIGN=CENTER>
                <TD VALIGN=CENTER>
                  <H2>Menu</H2>
                </TD>
              </TR>
              <TR>

                <TD ALIGN=center>
                  <A HREF="index.cgi">Home</A>
                </TD>
              </TR>
              <TR>
                <TD>
                  <A HREF="index.cgi?page=about">About</A>
                </TD>

              </TR>
              <TR>
                <TD>
                  <A HREF="http://ca.expasy.org/prosite/">ExPASy PROSITE</A>
                </TD>
              </TR>
              </TABLE>

              &nbsp;<br>&nbsp;
              <CENTER><IMG SRC="prosite.gif"></CENTER>
              <CENTER><IMG SRC="protein.gif"></CENTER>
            </TD>

            <TD WIDTH=80%>
              <TABLE WIDTH=90% ALIGN=LEFT VALIGN=TOP CLASS=dataSetTable CELLPADDING=10>
              <TR>
                <TD>
				  );
	 return $html_top;
}




# The dynamic middle panes, depending on the "state" the middle pane is different
sub get_html_main{
	 my $motifs_list_ref = shift;
	 my %motifs_list = %$motifs_list_ref;
	 my $html_main = start_form() . "\n";
	 $html_main.= qq(
                    Enter a FASTA sequence:
                    <BR>
						 );

	 $html_main.= textarea(-name=>"ta", 
								  -value=>"", 
								  -rows=>10, 
								  -cols=>85,
								  -wrap=>"physical" 
								  );

	 $html_main.= qq(
                    <BR>
                    <HR>
                    <BR>
                    PROSITE motif(s) to match:
                    <BR>

                    <TABLE ALIGN=LEFT VALIGN=TOP>
                    <TR>
                      <TD>
                      </TD>
                    </TR>
                    <TR>
                      <TD>
						  );
	 
	 my %lables = ('1'=>'ALL motifs from PROSITE.',
						'2'=>'Enter one PROSITE ID:'
						);
	 $html_main.= radio_group(-name=>'radio',
									  -values=>\%lables,				  
									  -linebreak=>'true',
									  -default=>'1'									  
									  );


	 $html_main.= qq(
						      <INPUT type=text name=txtID size=10>
                      </TD>
                    
                    </TR>
                    </TABLE>
                    <BR>
                    <BR>
                    <BR>
                    <BR>
                    <INPUT TYPE=submit value="SUBMIT" name=submit>
						  );

	 $html_main.= end_form() . "<BR><BR><BR><BR><BR>";
	 return $html_main;
}






sub get_html_about{
	 my $html_about = qq(
<h3>The Problem</h3><font size=3>
One of the biggest challenges in genome sequence analysis is to figure out what each predicted gene in a genome does. Since the majority of genes in most organisms are protein-coding, the challenge becomes guessing the function of the resultant protein encoded by a gene. Protein function prediction is done in a number of ways, but especially through the identification of conserved structural and sequence motifs.</font>
<br><br><h3>The Project</h3>
<font size=3>This tool searches the sequence the user inputs (FASTA format) for either one or all of the motif patterns in the PROSITE database. The location of each motif within the input sequence is displayed as text and a graphical image.<br><br>This project was completed as a requirement for the "Genomics & Bioinformatics" class. Fall 2006.</font>
<BR><BR><BR><BR><BR><BR><BR><BR><BR><BR><BR><BR>
								);
	 return $html_about;
}



# Results pane
sub get_html_results{
	 my $html_results;

	 # Parsing input sequence
	 my $input = param("ta");
	 my $stringfh = new IO::String($input);
	 my $seqio = Bio::SeqIO -> new(-fh => $stringfh, -format => 'fasta');
	 my ($id,$sequence);
	 while ( my $seq = $seqio->next_seq() ) {
		  $id = $seq->display_id();
		  $sequence = uc($seq->seq());
		  my $alph = $seq->alphabet();
		  # If the sequence is not amino acid then translate
		  if($alph eq "dna" or $alph eq "rna"){
				my $protein_seq =  $seq->translate();
				$sequence = $protein_seq->seq();
		  }
	 }

	 if($sequence eq ""){
		  $html_results.= qq(
									<h3>ERROR! Please enter a valid FASTA sequence.</h3>
									<BR><BR><BR><BR><BR><BR><BR><BR><BR><BR><BR><BR><BR><BR><BR><BR><BR><BR><BR><BR><BR><BR><BR><BR>
								);
		  return $html_results;
	 }

	 # Parse prosite file
	 my $motifs_ref = &prositeParser;
	 my %motifs = %$motifs_ref;
	 
	 # Motifs to match (user input)
	 my $matches_ref;
	 my %matches;

	 my $motif_to_match = param("radio");
	 if($motif_to_match eq "1"){
		  $matches_ref = &motifMatches(\%motifs,\$sequence,"all");
	 }
	 elsif($motif_to_match eq "2"){
		  my $id = &trim(param("txtID"));
		  $matches_ref = &motifMatches(\%motifs,\$sequence,$id);

	 }
	 elsif($motif_to_match eq "3"){
		  my $id = param("list");
		  $matches_ref = &motifMatches(\%motifs,\$sequence,$id);
	 }
	 #%matches = %$matches_ref;

	 # diplsay results
	 $html_results.= "<font name=courier size=3>" . &output($id,$sequence,$matches_ref,$motifs_ref) . "</font>\n<IMG SRC='../image/myimage.png'>";

	 $html_results.= qq(
							  <BR><BR><BR><BR><BR><BR><BR><BR><BR><BR><BR><BR><BR><BR><BR><BR><BR><BR><BR>
							  );
	 return $html_results;
}





# The static bottom of the html
sub get_html_bottom{
	 my $html_bottom = qq(
                </TD>
              </TR>
              </TABLE>
              <BR><BR><BR><BR><BR>
              <BR><BR><BR><BR><BR>
            </TD>

          </TR>
          </TABLE>
        </TD>
      </TR>
      </TABLE>
    </TD>
  </TR>
  <TR>
    <TD>

      &nbsp;
    </TD>
  </TR>
  </TABLE>);
	 return $html_bottom;
}




# Prints the output and generates the image
sub output{
	 # Getting input parameters
	 my $id = shift;
	 my $sequence = shift;
	 my $matches_ref = shift;
	 my %matches = %$matches_ref;
	 my $motifs_ref = shift;
	 
	 my $results;

	 # Creating a SeqFeature for the amino acid seqeunce
	 my $wholeseq = Bio::SeqFeature::Generic->new(-start=>1,-end=>length($sequence),
																 -display_name=>"$id");

	 # Creating a panel for the image
	 my $panel = Bio::Graphics::Panel->new(
														-length    => length($sequence),
														-key_style => 'between',
														-width     => 600,
														-pad_left  => 10,
														-pad_right => 10,
														);
	 # Adding an arrow glyph
	 $panel->add_track($wholeseq,
							 -glyph => 'arrow',
							 -bump => 0,
							 -double=>1,
							 -tick => 2);

	 # Adding a glyph to represent the complete sequence
	 $panel->add_track($wholeseq,
							 -glyph  => 'generic',
							 -bgcolor => 'blue',
							 -label  => 1,
							 );

	 # Determining if there are any motifs matched
	 # If there are none then add a glyph that displays a message 
	 my $display = "MOTIFS";
	 my $numb_of_motifs = keys(%matches);
	 if($numb_of_motifs<1){
		  $display = "ZERO MATCHES FOUND!";
	 }
	 $panel->add_track("",
							 -glyph  => 'generic',
							 -bgcolor => 'red',
							 -label  => "",
							 -key => "$display"								  
							 );

	 
	 # Go through the matched motifs
	 my ($key,$val);	 

	 $results="Found $numb_of_motifs matching motifs<BR><BR>";
	 		  
	 while(($key,$val) = each(%matches)){
		  my @outter_array;
		  my @positions = @$val;
		  my $desc = &getDesc($key, $motifs_ref);
		  my $pat = &getPattern($key, $motifs_ref);
		  # Print out motif information
		  $results.= "$key    $desc<BR>$pat<BR>";
		  # Print out the start and end indexes of each of the matches
		  foreach my $sub_pos (@positions){
				my @inner_array;
				my @start_end = split(/:/, $sub_pos);
				my $start = $start_end[0];
				my $end = $start_end[1];
				my $length = $end - $start;
				# Creating a two-element array of start and stops for each of the motifs
				push(@inner_array,$start_end[0]);
				push(@inner_array,$start_end[1]);
				push(@outter_array,\@inner_array);
				$results.= "Start: $start         End: $end          Motif: ". substr($sequence,$start,$length) ."<BR>";
		  }
		  $results.= "<BR>";

		  # Create a Feature for each motif that was matched
		  my $f = Bio::Graphics::Feature->new(-segments => \@outter_array,
														  -name     => $key,
														  -type     => 'motif',
														  -subtype  => '');
		  $panel->add_track($f,
								  -glyph  => 'generic',
								  -bgcolor => 'red',
								  -key => "$key: $desc"
								  );
		  
	 }

	 # Adding a track at the bottom of the image for display purposes
	 $panel->add_track("",
							 -glyph  => 'generic',
							 -bgcolor => 'red',
							 -label  => "",
							 );

	 # Write out image to file
	 open(OUT,">../image/myimage.png") or die;
	 binmode(OUT);
	 print OUT $panel->png;
	 close(OUT);
	 return $results;
}







# Parses a prosite file and stores the motif information
sub prositeParser{
	 	 
	 my $prosite_file = "prosite.dat";
	 open(IN, $prosite_file);

	 # Flag to see if an entry needs to be ignored (e.g when its a MATRIX)
	 my $ignore = "false";
	 my ($ID,$TYPE,$AC,$DE,$PA);
	 my %motifs;
	 
	 # Read in the sequence
	 while(<IN>){
		  chomp;
		  if($ignore eq "false"){
				if(/^ID/){ 
					 #clearing all variables
					 $ID ="";
					 $AC ="";
					 $DE ="";
					 $PA = "";
					 $TYPE = "";
	 
					 $ID = substr($_,5,length($_));
					 $TYPE = &getType($ID);
					 # Ignore entry if it is a MATRIX
					 if ($TYPE eq "MATRIX"){
						  $ignore="true";
					 }
				}
				elsif(/^AC/){
					 $AC=substr($_,5,length($_)-6);
				}
				elsif(/^DE/){
					 $DE=substr($_,5,length($_)-5);
				}
				elsif(/^PA/){
					 $PA.=substr($_,5,length($_)-6);
					 $PA=&regexp($PA); # Convert the patten to a regular expression
					 # Enter motif info into hash
					 $motifs{"$AC"} = "$ID|$DE|$PA";
				}
		  }
		  elsif($ignore eq "true" && /^DO/){
				# If the end of the entry that is being ignored; then stop ignoring (i.e its a new entry)
				$ignore = "false";
		  }
	 }

	 # Close the file
	 close(IN);

	 # output
	 return (\%motifs);
	 
}

# Print all the motifs and thier patterns
sub printMotifs{
	 my $motifs_ref= shift;
	 my %motifs = %$motifs_ref;
	 my ($key,$val);
	 while(($key,$val) = each(%motifs)){
		  print "$key: " . &getPattern($key) . "\n";
	 }
}

# Given Accesion number return the pattern
sub getPattern{
	 my $key = shift;
	 my $motifs_ref= shift;
	 my %motifs = %$motifs_ref;
	 my @vals;
	 if (exists  $motifs{ $key }){
		  @vals = split(/\|/, $motifs { $key });
	 }
	 return $vals[2];
}

# Given Accesion number return the desription of the motif
sub getDesc{
	 my $key = shift;
	 my $motifs_ref= shift;
	 my %motifs = %$motifs_ref;
	 my @vals;
	 if (exists  $motifs{ $key }){
		  @vals = split(/\|/, $motifs { $key });
	 }
	 return $vals[1];
}

# Given the ID returns the type of the entry
# e.g. PATTERN, RULE, MATRIX
sub getType{
	 my @split_ID =  split(/;/,$_[0]);
	 $split_ID[1] = trim($split_ID[1]);
	 return substr($split_ID[1],0,length($split_ID[1])-1);
}


# Removes leading & trailing spaces
# Original code obtained from http://www.somacon.com/p114.php
sub trim($){
	my $str = shift;
	# replace leading spaces
	$str =~ s/^\s+//;
	# replace trailing spaces
	$str =~ s/\s+$//;
	return $str;
}


# Given a PROSITE Pattern, this methods returns its 
# regular exprespssion equivilant
sub regexp{

######################################################################
#################### PROSITE PATTERN RULES ###########################
######################################################################
# 1 The standard IUPAC one-letter codes for the amino acids are used.
# 2 The symbol 'x' is used for a position where any amino acid is accepted.
# 3 Ambiguities are indicated by listing the acceptable amino acids for a 
#   given position, between square parentheses '[ ]'. For example: [ALT] 
#   stands for Ala or Leu or Thr.
# 4 Ambiguities are also indicated by listing between a pair of curly 
#   brackets '{ }' the amino acids that are not accepted at a given position. 
#   For example: {AM} stands for any amino acid except Ala and Met.
# 5 Each element in a pattern is separated from its neighbor by a '-'.
# 6 Repetition of an element of the pattern can be indicated by following that 
#   element with a numerical value or a numerical range between parenthesis. 
#   Examples: x(3) corresponds to x-x-x, x(2,4) corresponds to x-x or x-x-x or x-x-x-x.
# 7 When a pattern is restricted to either the N- or C-terminal of a sequence, 
#   that pattern either starts with a '<' symbol or respectively ends with a 
#   '>' symbol. In some rare cases (e.g. PS00267 or PS00539), '>' can also 
#   occur inside square brackets for the C-terminal element. 'F-[GSTV]-P-R-L-[G>]' 
#   means that either 'F-[GSTV]-P-R-L-G' or 'F-[GSTV]-P-R-L>' are considered.
# 8 A period ends the pattern.
######################################################################

	 my $pattern = shift;

	 # 5. Remove '-'
	 # Elements are separated by a '-' in a PROSITE pattern
	 # Not needed in a regular expression
	 $pattern =~(s/-//g);

	 # 2. Substitute the 'x' with '[ARNDCEQGHILKMFPSTWYV]'
	 # 'x' in the PROSITE pattern represents the position
	 # the amino acid is accepted
	 $pattern=~(s/x/[A-Z]/g);

	 # 4. Fix ambiguties NOT accepted
	 # {AM} stands for any amino acid except Ala and Met
	 # Needs to be converted to this format [^AM]
	 $pattern=~(s/{([A-Z]+)}/[^$1]/g);

	 # 6. Fix repetitions 
	 $pattern =~(s/\(([0-9]+)\)/{$1}/g);
	 $pattern =~(s/\((([0-9]+),([0-9]+))\)/{$1}/g);

	 # 7. Substitute '<' with '^'  (pattern at start of seq)
	 #    Susbstitute '>' with '$' (pattern at end of seq)
	 $pattern =~(s/\</^/);
	 $pattern =~(s/\>/\$/);

	 # 8. Ending period removed while parsing

	 return $pattern;
}


#  Loops through each motif from the PROSTE file and looks for occurences in 
#  input sequence
sub motifMatches{
	 my $motifs_ref = shift;	 
	 my %motifs = %$motifs_ref;
	 my $seq_ref = shift;
	 my $seq = $$seq_ref;
	 my $id = shift;
	 my %matches;

	 if($id eq "all"){
		  while ( my ($key,$value) = each(%motifs)){
				my $pattern = &getPattern($key,$motifs_ref);
				my @positions;
				# store all in array
				while($seq =~ /$pattern/g){
					 my $start=$-[0];
					 my $end = $+[0];
					 push(@positions,"$start:$end");
				}
				if(@positions>0){
					 $matches{"$key"} = \@positions;
				}
		  }
	 }
	 else{
		  if(exists  $motifs{ $id }){
				my $pattern = &getPattern($id,$motifs_ref);
				my @positions;
				# store all in array
				while($seq =~ /$pattern/g){
					 my $start=$-[0];
					 my $end = $+[0];
					 push(@positions,"$start:$end");
				}
				if(@positions>0){
					 $matches{"$id"} = \@positions;
				}
		  }

	 }
	 return (\%matches);
}
